const path = require("path")
const fs = require("fs")
const crypto = require('crypto')

const _ = require("lodash")
const slash = require('slash')
const url = require("url")
const urlencode = require('urlencode')
const queryString = require(`query-string`)
const cheerio = require("cheerio");
const chalk = require(`chalk`)

const { GraphQLJSON } = require(`gatsby/graphql`);

const highlight = require("gatsby-remark-prismjs/highlight-code.js");
//const { fluid } = require("gatsby-plugin-sharp");
const { fluid, stats, traceSVG } = require(`gatsby-plugin-sharp`)
const { createRemoteFileNode } = require(`gatsby-source-filesystem`)

const { replaceVariables, validateVariables, createMapReplacer } = require("./src/replace-variables.js");
const { removeCommonIndent } = require("./src/remove-common-indent.js");
const extractFragment = require("./src/extract-fragment.js");

// The transformation functions should be converted to plugins, but
// for now we keep them integrated to avoid proliferation of boilerplate.

const isRelativeUrl = url => {
  return url && !/^(https?:\/\/)|^\/|^#/i.test(url);
};

const notInPre = $ => (i, el) => $(el).parents("pre").length === 0;
const indexingAllowed = $ => (i, el) => $(el).data("indexing") !== "disabled";

/**
 * Calls the "gatsby-remark-prismjs" plugin's highlighting code and returns
 * the appropriate HTML.
 */
const highlightFragment = ($el, lang, code) => {
  const className = [$el.attr("class"), `language-${lang}`].filter(e => !!e).join(" ");

  return `<div class="gatsby-highlight">
    <pre class="${className}"><code data-language="${lang}">${highlight(lang, code, false)}</code></pre>
  </div>`;
};

/**
 * Embeds code from a separate file. Relative file path provided in the
 * data-embed attribute is resolved against the path of the file in which
 * the embed tag appears.
 */
const embedCode = ($, dir, variables, reporter) => {
  $("pre[data-embed]")
    .filter(notInPre($))
    .replaceWith((i, el) => {
      const $el = $(el);
      const declaredEmbed = $el.data("embed");
      const fragment = $el.data("fragment");
      const declaredLanguage = $el.data("language");
      const preserveIndent = $el.data("preserve-common-indent");

      // Replace variables in the path. We don't care about the semantics
      // here, it's up to the caller to ensure the path makes sense and is safe.
      const embed = replaceVariables(declaredEmbed, (name) => {
        let value = variables[name] || "";
        if (value.endsWith("/" || value.endsWith("\\"))) {
          return value.substring(0, value.length - 1);
        } else {
          return value;
        }
      });

      const embedAbsolute = path.resolve(dir, embed);
      if (!fs.existsSync(embedAbsolute)) {
        fail(`relative path ${embed}, resolved to ${embedAbsolute} does not exist.`);
        return "";
      }

      if (!fs.statSync(embedAbsolute).isFile()) {
        fail(`path ${embed} must point to a file.`);
        return "";
      }

      const ext = path.extname(embedAbsolute).substring(1).toLowerCase();
      const language = declaredLanguage || ext;

      const rawContent = fs.readFileSync(embedAbsolute, "utf8");
      let content;
      if (fragment) {
        try {
          content = extractFragment(rawContent, fragment);
        } catch (e) {
          fail(e);
          content = "";
        }
      } else {
        content = rawContent;
      }

      if (!preserveIndent) {
        content = removeCommonIndent(content);
      }

      // Ideally, we should just insert the raw contents and have it
      // highlighted in the dedicated code below, but cheerio has problems
      // serializing certain HTML tags (html, head, body), so we have to
      // highlight them here before cheerio has a chance to remove them.
      return highlightFragment($el, language, content);

      function fail(message) {
        const dot = message.endsWith("." ? "" : ".");
        reporter.warn(`Failed to embed content: ${message}${dot}`);
      }
    });

  return $;
};

/**
 * Highlights code snippets in place. The code reuses the highlighting
 * internals from the "gatsby-remark-prismjs" plugin.
 */
const highlightCode = $ => {
  // The $ wrapper is a mutable DOM/jQuery-like representation, so
  // we only need to select and modify links, returning the original $ reference.
  $("pre[data-language]")
    .replaceWith((i, el) => {
      const $el = $(el);
      const preserveIndent = $el.data("preserve-common-indent");

      let html = $el.html();
      if (!preserveIndent) {
        html = removeCommonIndent(html);
      }
      return highlightFragment($el, $el.data("language"), html);
    });
  return $;
};

/**
 * Rewrites local HTML links into links required by Gatsby. We do this to be
 * able to preserve working links in the source HTML files while feeding Gatsby
 * with absolute no-extension links it requires.
 */
const rewriteLinks = $ => {
  // The $ wrapper is a mutable DOM/jQuery-like representation, so
  // we only need to select and modify links, returning the original $ reference.
  $("a")
    .filter((i, link) => {
      const href = link.attribs.href;

      // Must end with .html
      return isRelativeUrl(href) && /\.html(#.*)?$/i.test(href);
    })
    .attr("href", (i, href) => {
      return "/" + href.replace(/\.html((#.*)?)$/i, "$1");
    });

  return $;
};

/**
 * Process images through "gatsby-plugin-sharp". This will copy and optimize
 * the image in multiple resolutions for different devices.
 */
const processImages = async ($, /*fileNodesByPath,*/ pathPrefix, createNode, getNodesByType, reporter, store, cache, variables, options) => {
  const $img = $("img").filter(notInPre($));


  //console.log('processImages OPTIONS >>> + ',options)

  // Collect images whose relative paths point at existing files.
  const imageNodesToProcess = [];
    const imageNodes = await Promise.resolve(
  $img.each( async (i, img) => {
    const src = urlencode(img.attribs.src);

      //variables.push(img)

      // download image and create a File node
      // with gatsby-transformer-sharp and gatsby-plugin-sharp
      // that node will become an ImageSharp
      if (!src)
          return

      const fileNode = await createRemoteFileNode({
          url: src,
          store,
          cache,
          createNode,
          createNodeId: id => `wp-image-sharp-${id}`,
          //ext: ".jpg",
      })


   //const fileNodesByPath = _.keyBy(getNodesByType("File"), n => n.relativePath);

    //if (_.has(fileNodesByPath, src)) {
    if (fileNode) {
      imageNodesToProcess.push(fileNode);
        if (options.verbose)
        reporter.success(`Image file [${fileNode.name}] FOUND for img src ${src}.`);
    } else {
      reporter.warn(`Image file not found for img src ${src}.`);
    }
  }))
  //.then((imggrab) => {
  //    console.log('here')
  //})

  // Process the images through the sharp plugin.
  const processed = await Promise.all(
    imageNodesToProcess.map(n =>
        fluid({
            file: n,
            args: options,
            reporter,
            cache
        })
    )
  );


  const processedByRelativePath = processed.reduce((map, fluid, i) => {
    map[imageNodesToProcess[i].url] = fluid;
    return map;
  }, {});


  // Replace the images in the HTML.
  $img
    .filter((i, img) => _.has(processedByRelativePath, img.attribs.src))
    .replaceWith((i, img) => {
        //console.log(img)
      const fluid = processedByRelativePath[img.attribs.src];
      const className = [img.attribs.class, "fluid"].filter(e => !!e).join(" ");
      const ratio = `${(1 / fluid.aspectRatio) * 100}%`;

      let imgResult

        // This allows subplugins to use Node APIs bound to `gatsby-transformer-remark`
        // to customize the GraphQL schema. This makes it possible for subplugins to
        // modify types owned by `gatsby-transformer-remark`.
        options.plugins.forEach(plugin => {
            const resolvedPlugin = require(plugin.resolve)
            if (typeof resolvedPlugin.addLazyLoad === `function`) {
                imgResult = resolvedPlugin.addLazyLoad(
                    img,
                    className,
                    fluid,
                    ratio,
                    plugin.pluginOptions
                )
            }

            //console.log('RESOLVED PLUGIN >>> ',plugin)
        })

        if (imgResult)
            return imgResult
        else
        return `<div style="position: relative">
        <span style="padding-bottom: ${ratio}; background-image: url('${fluid.base64}')" 
              class="fluid preview"> </span>
        <img class="${className}"
             alt="${img.attribs.alt || ""}"
             title="${img.attribs.title || ""}"
             src="${fluid.src}"
             srcSet="${fluid.srcSet}"
             sizes="${fluid.sizes}" />
      </div>`;



    });

  return $;
};

const anchorSvg = `<svg aria-hidden="true" height="16" viewBox="0 0 16 16" width="16"><path fill-rule="evenodd" d="M4 9h1v1H4c-1.5 0-3-1.69-3-3.5S2.55 3 4 3h4c1.45 0 3 1.69 3 3.5 0 1.41-.91 2.72-2 3.25V8.59c.58-.45 1-1.27 1-2.09C10 5.22 8.98 4 8 4H4c-.98 0-2 1.22-2 2.5S3 9 4 9zm9-3h-1v1h1c1 0 2 1.22 2 2.5S13.98 12 13 12H9c-.98 0-2-1.22-2-2.5 0-.83.42-1.64 1-2.09V6.25c-1.09.53-2 1.84-2 3.25C6 11.31 7.55 13 9 13h4c1.45 0 3-1.69 3-3.5S14.5 6 13 6z" /></svg>`;

const addSectionAnchors = $ => {
  $("section[id] > :header")
    .filter((i, el) => el.name !== "h1")
    .filter((i, el) => $(el).parents("pre[data-language]").length === 0) // don't process HTML inside pre
    .replaceWith((i, el) => {
      const $el = $(el);
      return `<${el.name}>
        <a class="anchor" href="#${$el.parent().attr("id")}" aria-hidden="true">${anchorSvg}</a>${$el.text()}
      </${el.name}>`;
    });
  return $;
};

const generateId = text => crypto.createHash('md5')
  .update(text).digest("hex").substring(8);

const makeUnique = (id, existing) => {
  let unique ;
  if (existing.has(id)) {
    let suffix = 0;
    do {
      unique = `${id}_${suffix}`;
      suffix++;
    } while (existing.has(unique));
  } else {
    unique = id;
  }
  existing.add(unique);
  return unique;
};

const setId = ($f, existing) => {
  $f.attr("id", makeUnique($f.attr("id") ||
    generateId(normalize($f.text())), existing));
};

const addIdsForIndexableFragments = $ => {
  const existing = new Set();

  forEachFullTextFragment($, $f => setId($f, existing));
  $(".warning, .info").each((i, e) => setId($(e), existing));
  return $;
};

/**
 * The cheerio HTML parser when emitting modified HTML globally replaces
 * certain characters (<>") with HTML entities. It also replaces quotes
 * with entities inside HTML tags, which breaks the rendering of highlighted
 * Javascript snippets.
 *
 * To work around this we switch off entity replacement in cheerio, but
 * then it also does not escape < inside tags, which also breaks rendering.
 * This method patches this problem by replacing < with entities in
 * the final serialized HTML.
 */
const fixClosingTagsInHighlightedCode = html => {
  return html.replace(/<span class="token punctuation"><(\/?)<\/span>/g,
    `<span class="token punctuation">&lt;$1</span>`)
    .replace(/<span class="token doctype"></g,
      `<span class="token doctype">&lt;`);
};

/**
 * Builds a table of contents JSON based on section nesting and headings.
 */
const createToc = $ => {
  return $("article > section[id]")
    .filter(notInPre($))
    .map(function asToc(i, e) {
      const $section = $(e);
      const $subsections = $section.children("section[id]");
      return {
        heading: $section.children(":header").eq(0).text(),
        anchor: $section.attr("id"),
        ...$subsections.length > 0 && { sections: $subsections.map(asToc).get() }
      };
    }).get();
};

const removeEmpty = a => {
  return a.filter(e => !!e).join(" ");
};

const forEachFullTextFragment = ($, cb) => {
  const elements = [ "p", "li"];
  const isIndexed = indexingAllowed($);

  elements.forEach(tag => {
    $(`${tag}`).each((i, e) => {
      const $e = $(e);
      // Don't index the first paragraph of figure
      // caption, it's indexed as a figure heading.
      if ($e.closest("figcaption").length > 0 && $e.is("p:first-child")) {
        return;
      }

      // Don't index if a child of an element with data-indexing="disabled".
      const withFlag = $e.closest("[data-indexing]");
      if (withFlag.length > 0 && !isIndexed(0, withFlag.get(0))) {
        return;
      }

      if ($e.parents("[data-marker]").length > 0 ||
        $e.find("[data-marker]").length > 0) {
        return;
      }
      cb($e);
      $e.attr("data-marker", "");
    });
  });

  $("[data-marker]").removeAttr("data-marker");
};

const getFigureCaption = $e => {
  const $caption = $e.find("figcaption");
  if ($caption.children().length > 0) {
    return normalize($caption.children().eq(0).text());
  } else {
    return normalize($caption.text());
  }
};

const headingExtractors = [
  {
    selector: "article",
    type: "heading",
    class: () => "section",
    text: $e => $e.children(":header").text()
  },
  {
    selector: "section[id]",
    type: "heading",
    class: () => "section",
    text: $e => $e.children(":header").text()
  },
  {
    selector: "code[id]",
    type: "code",
    class: () => "api",
    text: $e => $e.text()
  },
  {
    selector: "figure[id]",
    type: "figure",
    class: $e => {
      if ($e.find("img, picture").length > 0) {
        return "image";
      }
      if ($e.find("pre").length > 0) {
        return "example";
      }
      return "figure";
    },
    text: getFigureCaption
  },
  {
    selector: ".warning, .info",
    type: "heading",
    class: () => null,
    text: $e => $e.find("strong").eq(0).text()
  }
];

const normalize = t => {
  return t.trim().replace(/(\s|\n)+/g, " ");
};

const collectIndexableFragments = $ => {
  const isIndexed = indexingAllowed($);
  const fragments = [];
  if (!isIndexed(0, $("article"))) {
    return fragments;
  }

  const extractParents = ($e, includeCaption) => {
    const headings = $e.parents("section, article, .warning, .info")
      .children(":header, strong")
      .map((i, heading) => normalize($(heading).text())).get().reverse();

    // For paragraphs inside figure caption,
    // add figure heading to the list of parents.
    const $f = $e.closest("figure");
    if (includeCaption && $f.length > 0) {
      headings.push(getFigureCaption($f));
    }

    return headings;
  };

  headingExtractors.forEach(extractor => {
    $(extractor.selector)
      .filter(isIndexed)
      .each((i, e) => {
        const $e = $(e);
        fragments.push({
          text: normalize(extractor.text($e)),
          type: extractor.type,
          id: $e.attr("id") || "",
          parents: extractParents($e, false),
          class: removeEmpty([extractor.class($e), $e.attr("class")])
        });
      });
  });

  forEachFullTextFragment($, $f => {
    fragments.push({
      text: normalize($f.text().trim()),
      type: "paragraph",
      id: $f.attr("id"),
      parents: extractParents($f, true),
      class: removeEmpty([$f.attr("class")])
    })
  });

  return fragments;
};

// Gatsby API implementation
/*exports.onCreatePage = ({ page, actions }) => {
    const {createPage, deletePage} = actions

    //console.log('WP PAGE ===== ', page)

}*/

const onCreateNode = async ({
                              node,
                              actions,
                              loadNodeContent,
                              createNodeId,
                              createContentDigest,
                            }) => {
  const { createNode, createParentChildLink } = actions;

    //if (node.internal.type == `GraphQLSource`) {
    //    const rmods = await loadNodeContent(node)
    //    console.log(rmods)
   // }

  if (node.internal.type !== `WPGraphQL_Post`) {
    return;
  }

  //console.log('WP ===== ', node)

  const rawHtml = await loadNodeContent(content);
  let $ = cheerio.load(rawHtml, { decodeEntities: false });

  const htmlNode = {
    rawHtml: rawHtml,
    frontmatter: {
      id: node.name,
      title: normalize($("h1").eq(0).text())
    },

    id: createNodeId(`${node.id} >>> WPGraphQL_PostHTML`),
    children: [],
    parent: node.id,
    dir: node.dir,
    internal: {
      contentDigest: createContentDigest(rawHtml),
      type: "WPGraphQL_PostHtml",
    },
  };

  createNode(htmlNode);
  createParentChildLink({ parent: node, child: htmlNode });
};




const getSharpHtml = ({
                          node,
                          actions,
                          loadNodeContent,
                          createNodeId,
                          createContentDigest,
                          content,
                          pathPrefix,
                          reporter,
                          cache
                      },{ variables }) => {

    return {
        html: {
            type: "String",
            resolve: async (node) => {
                const fileNodesByPath = _.keyBy(getNodesByType("File"), n => n.relativePath);

                // For correct highlighting of HTML code, we need to disable
                // entity resolution in cheerio and then patch this in the
                // serialized HTML, see fixClosingTagsInHighlightedCode() below.
                let $ = cheerio.load(node.rawHtml, { decodeEntities: false });
                $ = await processImages($, fileNodesByPath, pathPrefix, reporter, cache);
                $ = rewriteLinks($);
                $ = addSectionAnchors($);
                $ = embedCode($, node.dir, variables, reporter);
                $ = highlightCode($);
                $ = addIdsForIndexableFragments($);

                let html = fixClosingTagsInHighlightedCode($.html("article"));
                html = replaceVariables(html, createMapReplacer(variables));

                return html;
            }
        },
        tableOfContents: {
            type: GraphQLJSON,
            resolve: (node) => {
                return createToc(cheerio.load(node.rawHtml));
            }
        },
        indexableFragments: {
            type: GraphQLJSON,
            resolve: (node) => {
                let $ = cheerio.load(node.rawHtml);
                $ = addIdsForIndexableFragments($);
                return collectIndexableFragments($);
            }
        }
    }

}

const getSharpHtml2 = async (
                                node,
                                actions,
                                loadNodeContent,
                                createNodeId,
                                createContentDigest,
                                content
                            ) => {
    const { createNode, createParentChildLink } = actions;

    //if (node.internal.type == `GraphQLSource`) {
    //    const rmods = await loadNodeContent(node)
    //    console.log(rmods)
    // }

    //if (node.internal.type !== `WPGraphQL_Post`) {
    //    return;
    //}

    //console.log('WP ===== ', node)

    const rawHtml = content;//loadNodeContent(content);
    let $ = cheerio.load(rawHtml, { decodeEntities: false });

    const htmlNode = {
        rawHtml: rawHtml,
        frontmatter: {
            id: node.name,
            title: normalize($("h1").eq(0).text())
        },

        id: createNodeId(`${node.id} >>> WPGraphQL_PostHTML`),
        children: [],
        parent: node.id,
        dir: node.dir,
        internal: {
            contentDigest: createContentDigest(rawHtml),
            type: "WPGraphQL_PostHtml",
        },
    };

    createNode(htmlNode);
    //createParentChildLink({ parent: node, child: htmlNode });
    return rawHtml
};

//###########################################################################
//# get Html Sharp
//###########################################################################
const getHtmlBeta = async (content, actions, createNodeId, getNodesByType, pathPrefix, store,cache, options, reporter, variables) => {

    const { createNode } = actions

    if (!variables)
        variables = []

    // For correct highlighting of HTML code, we need to disable
    // entity resolution in cheerio and then patch this in the
    // serialized HTML, see fixClosingTagsInHighlightedCode() below.
    let $ = cheerio.load(content, { decodeEntities: false });
    $ = await processImages($, pathPrefix, createNode, getNodesByType, reporter, store, cache, variables, options);
    $ = rewriteLinks($);
    $ = addSectionAnchors($);
    //$ = embedCode($, node.dir, variables, reporter);
    $ = highlightCode($);
    $ = addIdsForIndexableFragments($);

    let html = fixClosingTagsInHighlightedCode($.html());
    //var ss = createMapReplacer(variables)
    //reporter.info( $.html() )
    //html = replaceVariables(html, createMapReplacer(variables));


    return $(`body`).html();

}

const getHtml = async (content, actions, createNodeId, getNodesByType, store,cache, options, reporter) => {

    const { createNode } = actions

    // Takes a node and generates the needed images and then returns
    // the needed HTML replacement for the image
    const generateImagesAndUpdateNode = async function (
        node,
        resolve,
        inLink,
        overWrites = {}
    ) {
        // Check if this markdownNode has a File parent. This plugin
        // won't work if the image isn't hosted locally.
        const parentNode = node//getNode(node.parent)

        //console.info('parentNode     ', parentNode)

        //let imagePath = node.absolutePath
        //let imagePath =  slash(path.join(parentNode.dir, getImageInfo(node.url).url))
        let imagePath
        if (parentNode && parentNode.dir) {
            var parsedFilename = url.parse(getImageInfo(node.url).url);
            //console.log(path.posix.basename(parsedFilename.pathname));
            imagePath = slash(path.join(parentNode.dir, path.posix.basename(parsedFilename.pathname)))
        } else {
            return null
        }

        console.info('imagePath      ', imagePath)

        //const fileNodesByPath = _.keyBy(getNodesByType("File"), n => n.relativePath);

        //console.log('fileNodesByPath =============== ', fileNodesByPath.id)

        //const imageNode = {absolutePath: node.absolutePath}
        const imageNode = _.find([node], file => {
            if (file && file.absolutePath) {
                //console.info('FILE      ', file.absolutePath, imagePath)
                return file.absolutePath === imagePath
            }
            return null
        })


        //console.info('imageNode      ', _.find([node]))

        if (!imageNode || !imageNode.absolutePath) {
            return resolve()
        }

        let fluidResult = await fluid({
            file: imageNode,
            args: options,
            reporter,
            cache,
        })


        if (!fluidResult) {
            return resolve()
        }

        const originalImg = fluidResult.originalImg
        const fallbackSrc = fluidResult.src
        const srcSet = fluidResult.srcSet
        const presentationWidth = fluidResult.presentationWidth

        // Generate default alt tag
        const srcSplit = getImageInfo(node.url).url.split(`/`)
        const fileName = srcSplit[srcSplit.length - 1]
        const fileNameNoExt = fileName.replace(/\.[^/.]+$/, ``)
        const defaultAlt = fileNameNoExt.replace(/[^A-Z0-9]/gi, ` `)

        const alt = _.escape(
            overWrites.alt ? overWrites.alt : node.alt ? node.alt : defaultAlt
        )

        const title = node.title ? _.escape(node.title) : alt

        const loading = options.loading

        if (![`lazy`, `eager`, `auto`].includes(loading)) {
            reporter.warn(
                reporter.stripIndent(`
                        ${chalk.bold(loading)} is an invalid value for the ${chalk.bold(
                    `loading`
                )} option. Please pass one of "lazy", "eager" or "auto".
                      `)
            )
        }

        // Create our base image tag
        let imageTag = `
                          <img
                            class="${imageClass}"
                            alt="${alt}"
                            title="${title}"
                            src="${fallbackSrc}"
                            srcset="${srcSet}"
                            sizes="${fluidResult.sizes}"
                            loading="${loading}"
                          />
                        `.trim()

        // if options.withWebp is enabled, generate a webp version and change the image tag to a picture tag
        if (options.withWebp) {
            const webpFluidResult = await fluid({
                file: imageNode,
                args: _.defaults(
                    {toFormat: `WEBP`},
                    // override options if it's an object, otherwise just pass through defaults
                    options.withWebp === true ? {} : options.withWebp,
                    pluginOptions,
                    DEFAULT_OPTIONS
                ),
                reporter,
            })

            if (!webpFluidResult) {
                return resolve()
            }

            imageTag = `
                  <picture>
                    <source
                      srcset="${webpFluidResult.srcSet}"
                      sizes="${webpFluidResult.sizes}"
                      type="${webpFluidResult.srcSetType}"
                    />
                    <source
                      srcset="${srcSet}"
                      sizes="${fluidResult.sizes}"
                      type="${fluidResult.srcSetType}"
                    />
                    <img
                      class="${imageClass}"
                      src="${fallbackSrc}"
                      alt="${alt}"
                      title="${title}"
                      loading="${loading}"
                    />
                  </picture>
                  `.trim()
        }

        let placeholderImageData = fluidResult.base64

        // if options.tracedSVG is enabled generate the traced SVG and use that as the placeholder image
        if (options.tracedSVG) {
            let args = typeof options.tracedSVG === `object` ? options.tracedSVG : {}

            // Translate Potrace constants (e.g. TURNPOLICY_LEFT, COLOR_AUTO) to the values Potrace expects
            const {Potrace} = require(`potrace`)
            const argsKeys = Object.keys(args)
            args = argsKeys.reduce((result, key) => {
                const value = args[key]
                result[key] = Potrace.hasOwnProperty(value) ? Potrace[value] : value
                return result
            }, {})

            const tracedSVG = await traceSVG({
                file: imageNode,
                args,
                fileArgs: args,
                cache,
                reporter,
            })

            // Escape single quotes so the SVG data can be used in inline style attribute with single quotes
            placeholderImageData = tracedSVG.replace(/'/g, `\\'`)
        }

        const ratio = `${(1 / fluidResult.aspectRatio) * 100}%`

        const wrapperStyle =
            typeof options.wrapperStyle === `function`
                ? options.wrapperStyle(fluidResult)
                : options.wrapperStyle

        // Construct new image node w/ aspect ratio placeholder
        const imageCaption =
            options.showCaptions && getImageCaption(node, overWrites)

        let removeBgImage = false
        if (options.disableBgImageOnAlpha) {
            const imageStats = await stats({file: imageNode, reporter})
            if (imageStats && imageStats.isTransparent) removeBgImage = true
        }

        const bgImage = removeBgImage
            ? ``
            : ` background-image: url('${placeholderImageData}'); background-size: cover;`

        let rawHTML = `
                          <span
                            class="${imageBackgroundClass}"
                            style="padding-bottom: ${ratio}; position: relative; bottom: 0; left: 0;${bgImage} display: block;"
                          ></span>
                          ${imageTag}
                          `.trim()

        // Make linking to original image optional.
        if (!inLink && options.linkImagesToOriginal) {
            rawHTML = `
                          <a
                            class="gatsby-resp-image-link"
                            href="${originalImg}"
                            style="display: block"
                            target="_blank"
                            rel="noopener"
                          >
                            ${rawHTML}
                          </a>
                            `.trim()
        }

        rawHTML = `
                    <span
                      class="${imageWrapperClass}"
                      style="position: relative; display: block; margin-left: auto; margin-right: auto; ${
            imageCaption ? `` : wrapperStyle
            } max-width: ${presentationWidth}px;"
                    >
                      ${rawHTML}
                    </span>
                    `.trim()

        // Wrap in figure and use title as caption
        if (imageCaption) {
            rawHTML = `
                      <figure class="gatsby-resp-image-figure" style="${wrapperStyle}">
                        ${rawHTML}
                        <figcaption class="gatsby-resp-image-figcaption">${imageCaption}</figcaption>
                      </figure>
                          `.trim()
        }

        return rawHTML
    }



    const getImageInfo = uri => {
        const {url, query} = queryString.parseUrl(uri)
        return {
            ext: path
                .extname(url)
                .split(`.`)
                .pop(),
            url,
            query,
        }
    }

    const $ = cheerio.load(content, {xmlMode: true, decodeEntities: false})
    //let THEcontent = null

    let THEcontent = new Promise((resolve, reject) => {

        console.log('imageRefs ================ ', $(`img`).length)

        if ($(`img`).length === 0) {
            // No img tags
            return resolve()
        }

        let imageRefs = []
        $(`img`).each(async (idx, cImg) => {
            //imageRefs.push($(this))
            //reporter.success(`IMGL %%%%%%>> ${$(cImg).attr(`src`)} - ${idx}`)
        //})



        //for (let thisImg of imageRefs) {
            // Get the details we need.

            let formattedImgTag = {}
            formattedImgTag.url = $(cImg).attr(`src`)
            formattedImgTag.title = $(cImg).attr(`title`)
            formattedImgTag.alt = $(cImg).attr(`alt`)

            if (!formattedImgTag.url) {
                return resolve()
            }

            const fileType = getImageInfo(formattedImgTag.url).ext

            // Ignore gifs as we can't process them,
            // svgs as they are already responsive by definition
            if (
                !isRelativeUrl(formattedImgTag.url) &&
                fileType !== `gif` &&
                fileType !== `svg`
            ) {


                // download image and create a File node
                // with gatsby-transformer-sharp and gatsby-plugin-sharp
                // that node will become an ImageSharp
                const fileNode = await createRemoteFileNode({
                    url: formattedImgTag.url,
                    store,
                    cache,
                    createNode,
                    createNodeId: id => `wp-image-sharp-${id}`,
                })


                if (fileNode) {
                    // link File node to WpImage node
                    // at field image
                    //edge.node.image___NODE = fileNode.id
                    //console.log(generateImagesAndUpdateNode(fileNode))

                    //var _html = fileNode.then((fnode) => generateImagesAndUpdateNode(fnode,resolve).then((result) => {
                    var _html = generateImagesAndUpdateNode(fileNode,resolve).then((result) => {

                        //reporter.success(`rawHTML %%%%%% ${fileNode.url} -- ${result}`)
                        //reporter.success(`rawHTML %%%%%% ${rawHTML}`)

                        var _thisImg = $(result)

                        //reporter.info(`thisImg %%%%%%$$$$$$$$$$$ ${_thisImg}`)
                        //reporter.info(`thisImg %%%%%%$$$$$$$$$$$ ${$(cImg).replaceWith(_thisImg)}`)
                        //reporter.info(`thisImg %%%%%%$$$$$$$$$$$ ${_thisImg.replaceWith($(cImg))}`)

                        //$(cImg).replaceWith(_thisImg)
                        //reporter.success(`thisImg replaceWith %%%%%%==== ${$.html()}`)
                        return $.html()
                    })

                    //console.log('END ================== ', _html)
                    return resolve(_html)

                } else {
                    return resolve()
                }
            }
        })

         //console.log('$$$$$$$$$$$$$$$$$$$$$$$$$$$$$ HTML @@@@@@@@@@@@@@@@@@@@@@@@@@ --- ', $(`body`).html())
        //edge.node.content = $(`body`).html()
        //return 'TRUE'
    }).then(async (content) => {
        //console.log(content)
        return content
    })

    return await THEcontent

}


exports.createSchemaCustomization = ({ actions, reporter}, pluginOptions = {}) => {
    actions.createResolverContext({ getHtmlBeta, getHtml })

    const { plugins = [] } = pluginOptions
    //console.log('Plugin Options >>> ',pluginOptions)

    // This allows subplugins to use Node APIs bound to `gatsby-transformer-remark`
    // to customize the GraphQL schema. This makes it possible for subplugins to
    // modify types owned by `gatsby-transformer-remark`.
    plugins.forEach(plugin => {
        const resolvedPlugin = require(plugin.resolve)
        if (typeof resolvedPlugin.createSchemaCustomization === `function`) {
            resolvedPlugin.createSchemaCustomization(
                //nodeApiArgs,
                plugin.pluginOptions
            )
        }

        console.log('PLUGIN >>> ',plugin)
    })

    //actions.createResolverContext({ getHtml })
    //actions.createResolverContext({ getSharpHtml })
}
// The context value can then be accessed in any field resolver like this:
/*exports.createSchemaCustomization = ({ actions, createNodeId, createContentDigest, schema, reporter, pluginOptions = {} }) => {
// The context value can then be accessed in any field resolver like this:

    const { plugins = [] } = pluginOptions
    reporter.info('Plugin Options >>> ',pluginOptions)

    // This allows subplugins to use Node APIs bound to `gatsby-transformer-remark`
    // to customize the GraphQL schema. This makes it possible for subplugins to
    // modify types owned by `gatsby-transformer-remark`.
    plugins.forEach(plugin => {
        const resolvedPlugin = require(plugin.resolve)
        if (typeof resolvedPlugin.createSchemaCustomization === `function`) {
            resolvedPlugin.createSchemaCustomization(
                //nodeApiArgs,
                plugin.pluginOptions
            )
        }

        reporter.info('PLUGIN >>> ',plugin)
    })
}*/


/*exports.createSchemaCustomization = ({ actions, createNodeId, createContentDigest, schema }) => {
    const { createTypes, createNode, createParentChildLink } = actions;
    createTypes(schema.buildObjectType({
        name: 'WPGraphQL_PostHTML',
        interfaces: ['Node'],
        fields: {
            content: {
                type: 'String!',
                async resolve(source, args, context, info) {
                    console.log( context, '==================================')
                    //const processed = await context.transformerRemark.getHtml(source.internal.contents)
                    //return processed.contents
                }
            }
        }
    }))
}*/

const {
    DEFAULT_OPTIONS,
    imageClass,
    imageBackgroundClass,
    imageWrapperClass,
} = require(`./src/constants`)

const createResolvers = ({
                             actions,
                             cache,
                             pathPrefix,
                             getNodesByType,
                             createNodeId,
                             createContentDigest,
                             loadNodeContent,
                             createResolvers,
                             store,
                             reporter,
                             variables
                         }, pluginOptions) => {

    const { plugins = [] } = pluginOptions
    const options = _.defaults(pluginOptions, { pathPrefix }, DEFAULT_OPTIONS)

    //console.log('RESOLVE PLUGINS >>>> ',  plugins)
    //console.log('Plugin Options >>> ',pluginOptions)

    const resolvers = {
        WPGraphQL_Post: {
            /*fullName: {
                resolve: (source, args, context, info) => {
                    return source.title + source.author
                }
            }*/
            content: {
                type: [`String`],
                //type: [`Node`],
                resolve: async (source, args, context, info) => {
                    //console.log( source)
                    let newcontent = source.content
                    let renderSharpcontent = await context.sekmetGatsbyTransformerWpcontent.getHtmlBeta( newcontent, actions, createNodeId, getNodesByType, pathPrefix, store, cache, options, reporter, variables )
                    /*let renderSharpcontent = await context.sekmetGatsbyTransformerWpcontent.getSharpHtml(
                        source,
                        actions,
                        loadNodeContent,
                        createNodeId,
                        createContentDigest,
                        newcontent,
                        pathPrefix,
                        reporter,
                        cache

                    )*/
                    //reporter.success('renderSharpcontent =====> ', renderSharpcontent)
                    ///*const pageDeps = context.nodeMode.trackPageDependencies
                    /*const content_sharp = context.nodeModel.getNodeById({
                        id: source.id,
                    })*/
                    return renderSharpcontent ? [renderSharpcontent] : [newcontent]
                    //return renderSharpcontent
                }
            },
        },

    }

    createResolvers(resolvers)
}

const setFieldsOnGraphQLNodeType = ({ type, getNodesByType, reporter, cache, pathPrefix }, { variables }) => {


    console.log('TYPE ============ ', type.name)

  if (type.name === "WPGraphQL_Post") {
    return {
      html: {
        type: "String",
        resolve: async (node) => {
          const fileNodesByPath = _.keyBy(getNodesByType("File"), n => n.relativePath);

          // For correct highlighting of HTML code, we need to disable
          // entity resolution in cheerio and then patch this in the
          // serialized HTML, see fixClosingTagsInHighlightedCode() below.
          let $ = cheerio.load(node.rawHtml, { decodeEntities: false });
          $ = await processImages($, fileNodesByPath, pathPrefix, reporter, cache);
          $ = rewriteLinks($);
          $ = addSectionAnchors($);
          $ = embedCode($, node.dir, variables, reporter);
          $ = highlightCode($);
          $ = addIdsForIndexableFragments($);

          let html = fixClosingTagsInHighlightedCode($.html("article"));
          html = replaceVariables(html, createMapReplacer(variables));

          return html;
        }
      },
      tableOfContents: {
        type: GraphQLJSON,
        resolve: (node) => {
          return createToc(cheerio.load(node.rawHtml));
        }
      },
      indexableFragments: {
        type: GraphQLJSON,
        resolve: (node) => {
          let $ = cheerio.load(node.rawHtml);
          $ = addIdsForIndexableFragments($);
          return collectIndexableFragments($);
        }
      }
    }
  }
};

const onPreExtractQueries = async ({ store, getNodesByType }) => {
    const program = store.getState().program

    // Check if there are any ImageSharp nodes. If so add fragments for ImageSharp.
    // The fragment will cause an error if there are no ImageSharp nodes.
    if (getNodesByType(`GraphQLSource`).length == 0) {
        return
    }

    //console.log(getNodesByType(`GraphQLSource`))

    // We have ImageSharp nodes so let's add our fragments to .cache/fragments.
    /*await fs.copy(
        require.resolve(`gatsby-transformer-sharp/src/fragments.js`),
        `${program.directory}/.cache/fragments/image-sharp-fragments.js`
    )*/
}

const onPreBootstrap = ({ reporter }, { variables }) => {
  try {
      ///console.log('variables ============ ', variables)
    validateVariables(variables);
  } catch (e) {
    reporter.panic(e);
  }
};
exports.createResolvers = createResolvers;
exports.onPreBootstrap = onPreBootstrap;
exports.onCreateNode = onCreateNode;
//exports.setFieldsOnGraphQLNodeType = setFieldsOnGraphQLNodeType;